package pacman;

import pacman.entries.ghosts.ChaserGhosts;
import pacman.entries.ghosts.LineGhosts;
import pacman.entries.ghosts.RandomGhosts;
import pacman.entries.ghosts.StandardGhosts;
import pacman.game.Constants;

import java.awt.Image;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.nio.Buffer;
import java.text.DecimalFormat;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.EnumMap;
import java.util.HashMap;
import java.util.Queue;
import java.util.Random;

import javax.imageio.ImageIO;


import pacman.entries.ghosts.*;
import pacman.entries.pacman.BasicRLPacMan;
import pacman.entries.pacman.CustomFeatureSet;
import pacman.entries.pacman.CustomFeatureSetV2;
import pacman.entries.pacman.DepthFeatureSet;
import pacman.entries.pacman.FeatureSet;
import pacman.entries.pacman.QPacMan;
import pacman.entries.pacman.RLPacMan;
import pacman.entries.pacman.SarsaPacMan;
import pacman.game.Game;
import pacman.game.GameView;
import pacman.game.Constants.GHOST;
import pacman.game.Constants.MOVE;
import pacman.game.internal.Maze;
import pacman.utils.DataFile;
import pacman.utils.LearningCurve;
import pacman.utils.Stats;

public class RunTest {
	
	//options for features include depth or custom, options for algoirithms are S or Q
	//i.e., customQ, customS, depthS, depthQ are the legal values for STUDENT
	public static String STUDENT = "depthS"; // Student feature set and algorithm
		
		
	public static int LENGTH = 500; // 10; // 100; // Points per curve
	public static int TEST = 1; //30; // Test episodes per point
	public static int TRAIN = 1; //10 // Train episodes per point

	public static Random rng = new Random();
		
	// Ghost "policies"
	public static StandardGhosts ghostsS = new StandardGhosts();
	public static RandomGhosts ghostsR = new RandomGhosts();
	public static ChaserGhosts ghostsC = new ChaserGhosts();
	public static LineGhosts ghostsL = new LineGhosts();
	    
	public static int ghostType = 0;		// 0 = RandomGhosts, 1 = StandardGhosts, 2 = ChaserGhosts
	    
	// Task parameters
	public static int mazeNum = 0;
	// Controls how fast ghosts move when pacman has eaten power pill. Lower numbers = slower (1 = frozen). 
	public static int ghostSlowdown = 2;
	public static int nGhosts = 4;
	    
	public static Constants defaultConstants;
	    
		/**
		 * Run experiments.
		 */
	public static void main(String[] args) {
		demo();
			
	}
	

	public static void demo(){
		defaultConstants = new Constants();
		
		//train params
		LENGTH = 100; //how many train episodes
		TEST = 1;
				
		mazeNum = 3; //which maze;
		ghostType = 1; //which of three types of ghosts
		ghostSlowdown = 2; //ghost speed when edible
		nGhosts = 4; //how many ghosts
		
		defaultConstants.MAZE_NUM = mazeNum; //which maze;
		defaultConstants.GHOST_TYPE = ghostType; //which of three types of ghosts
		defaultConstants.GHOST_SPEED_REDUCTION = ghostSlowdown; //ghost speed
		defaultConstants.NUM_GHOSTS = nGhosts; //how many ghosts
		//defaultConstants.nodeNames[4]="mymaze";
		//defaultConstants.distNames[4]="dmymaze";

		//increase delay for watching
		defaultConstants.DELAY = defaultConstants.DELAY*5;
				
		//train
		String demo_name = "independent_temp";
		train(demo_name,0,true,false);
				
	}
	
	/** Set up a learner. */
	public static RLPacMan create(String learner) {
		
		//create feature set based on constant STUDENT
		FeatureSet studentProto = STUDENT.startsWith("custom") ? new CustomFeatureSet() : new DepthFeatureSet();

		// Lone student
		if (learner.startsWith("independent")) {
			return STUDENT.endsWith("S") ? new SarsaPacMan(studentProto) : new QPacMan(studentProto);
		}
		
		return null;
	}
	
	public static String generateDirectoryName(String learner, 
			int mazeNum_i, int ghostType_i, 
			int ghostSlowdown_i, int nGhosts){
		return new String(learner+"_maze"+mazeNum_i+"_type"+ghostType_i+"_slow"+ghostSlowdown_i+"_ghosts"+nGhosts+"");
	}
	
	public static void episode(RLPacMan pacman) {

		Game game = new Game(rng.nextLong(), defaultConstants);
		pacman.startEpisode(game, false);

		while(!game.gameOver()) {
			game.advanceGame(pacman.getMove(game.copy(), -1), getGhostMove(game, game.constants.GHOST_TYPE));				
			pacman.processStep(game);
		}
	}
	

	private static EnumMap<GHOST, MOVE> getGhostMove(Game game, int ghostType){
		
		switch(ghostType){
			case 0:
				return ghostsR.getMove(game.copy(), -1);
			case 1:
				return ghostsS.getMove(game.copy(), -1);
			case 2:
				return ghostsC.getMove(game.copy(), -1);
			case 3:
				return ghostsL.getMove(game.copy(), -1);
			default:
				System.err.println("INVALID GHOST TYPE");
				System.exit(1);
				return null;				// Compilier is complaining... 
		}
	}
	
	/** Estimate the current performance of a learner. */
// Width is the number of episodes to test for (i.e. variable TEST above)	
	public static double[] evaluate(RLPacMan pacman, int width) {
		
		double[] scoreAndTime = new double[2];		
		double sumScore = 0;
		double sumSteps = 0;
		
// For each test episode		
		for(int i=0; i<width; i++) {
// Create and start a new game			
			Game game = new Game(rng.nextLong(), defaultConstants);
			pacman.startEpisode(game, true);
			while(!game.gameOver()) {
// getMove in pacman learner returns the lastAction	
// This will then recompute the next move to make.
				game.advanceGame(pacman.getMove(game.copy(), -1), getGhostMove(game, game.constants.GHOST_TYPE));			
				pacman.processStep(game);
			}
			
			sumScore += game.getScore();
			sumSteps += game.getTotalTime();
			//System.out.println("time taken: " + game.getTotalTime());
		}

		scoreAndTime[0] = sumScore/width;
		scoreAndTime[1] = sumSteps/width;
		
		
		return scoreAndTime;
	}
	

	/** Observe a learner play a game. */
	public static void watch(RLPacMan pacman, boolean destroyWindow) {
		
		
		Game game=new Game(rng.nextLong(), defaultConstants);
		
		pacman.startEpisode(game, true);
		GameView gv=new GameView(game).showGame();
		
		

		while(!game.gameOver()) {
			
			game.advanceGame(pacman.getMove(game.copy(), -1), getGhostMove(game, game.constants.GHOST_TYPE));	
			pacman.processStep(game);
			
			try{Thread.sleep(defaultConstants.DELAY);}catch(Exception e){}
			gv.repaint();
			Image img_i = gv.createImage(100, 100);
			BufferedImage bi = (BufferedImage)img_i;
			File f = new File("./output.png");
			try {
				ImageIO.write(bi, "png", f);
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		
		if (destroyWindow){
			gv.setVisible(false);
			gv.setEnabled(false);
			gv.getFrame().dispose();
		}
	}
	
	/** Generate learning curves. */
	public static void train(String learner, int start, boolean watchAtEnd, boolean watchDuring) {
		
		String directory = generateDirectoryName(learner,mazeNum,ghostType,ghostSlowdown,nGhosts);
		
		
		// This just sets up the student or teacher with the specified feature set and learning algorithm
		RLPacMan pacman = create(learner);
			
		// First point
		double[] initialData = pacman.episodeData();		// Starts off empty
			
		// evaluate the random policy
		double [] eval_result = evaluate(pacman, TEST);
		
		System.out.println("score at start = "+eval_result[0]);
			
		double initialScore = eval_result[0];		// This returns the average score over TEST number of games
		double initialTime = eval_result[1];		// This returns the average score over TEST number of games
			
		// Rest of the points
		int num_train_episodes = 0;
		
		// for each train episode...
		for (int x=1; x<=LENGTH; x++) {
			double[] data = new double[initialData.length];
				
			
			//make PacMan play an episode
			episode(pacman);	
			num_train_episodes++;
					
					
			double[] episodeData = pacman.episodeData();
			for (int d=0; d<data.length; d++)
				data[d] += episodeData[d];
			
				
				
			// now evaluate 	
			double [] eval_result2 =evaluate(pacman, TEST) ;
			double score = eval_result2[0];		// This returns the average score over TEST number of games
			double time = eval_result2[1];		// This returns the average number of game ssteps over TEST number of games
		
			//System.out.println("e"+num_train_episodes+" score:\t"+score+" ");
			
			System.out.println(""+num_train_episodes+","+score);
			
			if (watchDuring){
				((SarsaPacMan)pacman).debug=false;
				watch(pacman,true);
				((SarsaPacMan)pacman).debug=false;
			}
			
		}
		
		if (watchAtEnd)
			watch(pacman, false);
		
		
		System.out.println("Done.");
	}

}